from abc import ABCMeta, abstractmethod
from typing import Dict, List

import napalm.base.models as napalm_models
from . import models as umnet_models


class AbstractUMnetNapalm(metaclass=ABCMeta):
    """
    Abstract definition of umnet_napalm class.
    Some of these methods are expected to be inherited from
    NAPALM. Extensions to NAPALM are defined in vendor-specific child classes

    """

    @abstractmethod
    def __init__(self, hostname: str) -> str:
        """at minimum we need a hostname for mock"""
        self.hostname = hostname

    @abstractmethod
    def get_facts(self) -> napalm_models.FactsDict:
        """get basic device facts"""

    @abstractmethod
    def get_lldp_neighbors(self) -> Dict[str, List[napalm_models.LLDPNeighborDict]]:
        """get lldp neighbors"""

    @abstractmethod
    def get_ip_interfaces(self) -> List[umnet_models.IPInterfaceDict]:
        """get IP interface information"""

    @abstractmethod
    def get_arp_table(self) -> List[napalm_models.ARPTableDict]:
        """get IP interface information"""

    @abstractmethod
    def get_active_routes(self) -> List[umnet_models.RouteDict]:
        """get active routes"""

    @abstractmethod
    def get_mpls_switching(self) -> List[umnet_models.MPLSDict]:
        """get mpls switching (the mpls forwarding table)"""

    @abstractmethod
    def get_vni_information(self) -> List[umnet_models.VNIDict]:
        """get vni to vlan and VRF mapping"""

    @abstractmethod
    def get_inventory(self) -> List[umnet_models.InventoryDict]:
        """get inventory items"""



from .nxos import NXOS
from .ios import IOS
from .junos import Junos
from .panos import PANOS
from .mock import MockNXOS, MockIOS, MockJunos

PLATFORM_MAP = {"ios": IOS, "nxos_ssh": NXOS, "junos": Junos, "panos": PANOS}
MOCK_PLATFORM_MAP = {
    "ios": MockIOS,
    "nxos": MockNXOS,
    "junos": MockJunos,
}


def get_network_driver(platform: str):
    """
    Returns network driver based on platform string.
    """
    for valid_platform, driver in PLATFORM_MAP.items():
        if valid_platform == platform:
            return driver

    raise NotImplementedError(f"Unsupported platform {platform}")


def get_mock_driver(platform: str):
    """
    Returns mocked driver for testing
    """
    for valid_platform, driver in MOCK_PLATFORM_MAP.items():
        if valid_platform == platform:
            return driver

    raise NotImplementedError(f"Unsupported mock platform {platform}")



# # pylint: disable=no-member
from typing import List, Union
import re
from ipaddress import ip_network

from napalm.junos import JunOSDriver
from napalm.base.models import ARPTableDict

from lxml import etree

from ..base import UMnetNapalmError, UMnetNapalm
from ..models import RouteDict, MPLSDict, InventoryDict
from . import junos_views
from ..utils import age_to_datetime


XCVR_DESC_TO_INTERFACE_PREFIX = {
    r"QSFP": "et-",
    r"SFP\+": "xe-",
    r"SFP-": "ge-",
}

DEFAULT_ARP_AGING_TIMER = 1200

class Junos(UMnetNapalm, JunOSDriver):
    """
    Junos parser
    """

    # interfaces to ignore when doing 'show ip interfaces'
    IGNORE_INTERFACES = [
        "bme",
        "jsrv",
        "em1",
        "em0",
    ]

    # tables to ignore when doing 'show route'
    IGNORE_ROUTE_TABLES = [
        "bgp.l3vpn.0",
        "bgp.l3vpn-inet6.0",
        "inet.3",
        "inet6.3",
        "iso.0",
        "mpls.0",
        ":vxlan.inet.0",
        "__default_evpn__.evpn.0",
        "default-switch.evpn.0",
    ]

    # mapping of inventory name regex to their netbox types
    INVENTORY_TO_TYPE = {
        r"Fan Tray": "fan",
        r"(PIC|MIC|FPC)": "linecard",
        r"Xcvr": "optic",
        r"(PEM|Power Supply)": "psu",
        r"Routing Engine": "re",
        # don't care about these inventory items
        r"(CPU|PDM|FPM|Midplane|CB)": None,
    }

    def __init__(self, hostname:str, username:str, password:str, **kwargs):
        """
        overriding the napalm driver init class so we can keep internal state
        """
        super().__init__(hostname, username, password, **kwargs)
        self._config = None
        self.arp_aging_timer = DEFAULT_ARP_AGING_TIMER



    def _get_junos_inventory_type(self, name: str, model_number: str) -> str:
        """
        Maps the name and part number of the Junos inventory item to its type
        """
        # "Xcvr" parts are always optics
        if name.startswith("Xcvr"):
            return "optic"

        # non-optic parts should always have a model number, if not
        # we don't care about them
        if not (model_number):
            return None

        # EX switches and virtual chassis save their vc members and uplink modules
        # as FPC X and PIC X - we want to classify those correctly
        if re.match(r"EX[234]", model_number):
            if re.search(r"FPC \d+", name):
                return "stack-member"

            if re.search(r"PIC \d+", name):
                return None

        # uplink modules are also saved under PIC X, must classify based
        # on their model number
        if re.match(r"EX-UM", model_number):
            return "uplink-module"

        # otherwise we want to pattern-match based on the INVENTORY_TO_TYPE
        # dictionary
        return self._get_inventory_type(name)

    def _get_xcvr_interface_prefix(self, xcvr_desc: str) -> str:
        """
        Maps xcvr description from "show chassis hardware" to
        interface prefix
        """
        for pattern, prefix in XCVR_DESC_TO_INTERFACE_PREFIX.items():
            if re.match(pattern, xcvr_desc):
                return prefix

        raise UMnetNapalmError(f"{self.hostname}: Unknown xcvr type {xcvr_desc}")

    def _get_mod_number(self, mod_name: str) -> str:
        """
        Strips out "FPC|MIC|PIC" from a module name seen in "show chassis hardware"
        """
        return re.sub(r"(MIC|PIC|FPC) ", "", mod_name)

    def _send_command(self, view: str) -> str:
        """
        Uses junos views to send an RPC command and gets the results
        """
        view = getattr(junos_views, view)
        result = view(self.device)  # pylint: disable=no-member
        return result.get()

    def get_config(self):
        """
        If we run our nornir getters right we can pull the arp aging timer
        out of the config
        """
        self._config = super().get_config()
        return self._config
    
    def get_active_routes(self) -> List[RouteDict]:
        """
        Parses rpc call show route detail
        """
        result = self._send_command("junos_route_table")

        output = []

        for route in dict(result).values():
            if route.vrf in self.IGNORE_ROUTE_TABLES:
                continue

            prefix = ip_network(route.prefix)

            if route.vrf in ["inet.0", "inet6.0"]:
                vrf = "default"
            else:
                vrf = route.vrf.split(".")[0]

            # also skip link local prefixes
            if prefix.is_link_local:
                continue

            for entry in dict(route.rt_entry).values():
                for next_hop in dict(entry.next_hop).values():
                    if not next_hop.selected_next_hop:
                        continue

                    if entry.learned_from:
                        learned_from = entry.learned_from
                    elif next_hop.nh_ip:
                        learned_from = next_hop.nh_ip
                    elif entry.protocol in ["Direct", "Local", "Static"]:
                        learned_from = "self"
                    else:
                        raise UMnetNapalmError(
                            f"Could not determine learned from for {route}"
                        )

                    output.append(
                        RouteDict(
                            vrf=vrf,
                            prefix=prefix,
                            nh_interface=next_hop.nh_interface,
                            learned_from=learned_from,
                            protocol=entry.protocol,
                            age=age_to_datetime(entry.age),
                            nh_ip=next_hop.nh_ip,
                            mpls_label=self._parse_mpls_label(next_hop.mpls_label),
                            # not supporting junos VXLAN
                            vxlan_vni=None,
                            vxlan_endpoint=None,
                        )
                    )

        return output

    def get_inventory(self) -> List[InventoryDict]:
        """
        parses the get-chassis-inventory RPC call, which maps to "show chassis hardware"
        """
        result = self._send_command("junos_inventory")

        output = []
        for chassis in dict(result).values():

            # note that the chassis model and s/n are at this level, but
            # that doesn't count as an 'inventory item' so we're ignoring it

            # saving modules, sub-modules and sub-sub modules
            for module in dict(chassis.modules).values():
                self._save_inventory_item(output, module)

                for sub_module in dict(module.sub_modules).values():
                    self._save_inventory_item(output, sub_module, parent=module.name)

                    for sub_sub_module in dict(sub_module.sub_sub_modules).values():
                        self._save_inventory_item(
                            output,
                            sub_sub_module,
                            parent=sub_module.name,
                            grandparent=module.name,
                        )

        return output

    def _save_inventory_item(
        self, output: list, item: any, parent: str = None, grandparent: str = None
    ) -> bool:
        """
        Extracts data from a particular inventory item object.
        Returns whether we care aobut this item or not (so we know whether or not)
        to loop over its children
        """
        # skip builtin types, or parts without model numbers that aren't transcievers
        if item.part_number == "BUILTIN":
            return False

        # get inventory type based on item name and P/N
        inv_type = self._get_junos_inventory_type(item.name, item.model_number)

        if not inv_type:
            return False

        # for transcievers, change the name from Xcvr X to be the junos interface name
        # note that we expect transcievers to always be at the sub-sub module level
        # and thus to have a grandparent and parent
        m = re.search(r"Xcvr (\d+)", item.name)
        if m:
            if not (grandparent or parent):
                raise UMnetNapalmError(
                    f"{self.hostname}: No MIC and PIC found for {item.name}"
                )
            prefix = self._get_xcvr_interface_prefix(item.description)

            item_name = f"{prefix}{self._get_mod_number(grandparent)}/{self._get_mod_number(parent)}/{m.group(1)}"
        else:
            item_name = item.name

        output.append(
            {
                "type": inv_type,
                "name": item_name,
                "part_number": self._get_inventory_part_number(item),
                "serial_number": item.serial_number,
                "parent": parent,
            }
        )

        return True

    def _get_inventory_part_number(self, item: any) -> str:
        """
        Extracts the part number from an inventory item
        which, depending on the specific item, is stashed in
        the part number, model number, or description field
        """
        if item.model_number:
            return item.model_number
        return item.description

    def get_mpls_switching(self) -> List[MPLSDict]:
        """
        Parses 'show route active-path table mpls.0'
        """
        result = self._send_command("junos_mpls_table")

        output = []
        for label in dict(result).values():
            # in our network we can skip these 'label is not at the top
            # of the stack' entries - they're redundant
            # also labels 0-2 are junos-internal
            if "(S=0)" in label.prefix or label.prefix in ["0", "1", "2"]:
                continue

            for entry in dict(label.rt_entry).values():
                for next_hop in dict(entry.next_hop).values():

                    # next hop table is either in 'nh-table-receive' or 'nh-table'
                    nh_table = (
                        next_hop.nh_table
                        if next_hop.nh_table
                        else next_hop.nh_table_receive
                    )

                    (nh_interface, out_label, aggregate) = self._process_mpls_nh(
                        next_hop.nh_interface, next_hop.mpls_label, nh_table
                    )

                    # 'fec' and 'rd' aren't in the Junos mpls.0 table - this is a Cisco
                    # thing from "show mpls switching"
                    output.append(
                        {
                            "in_label": label.prefix,
                            "out_label": out_label,
                            "nh_interface": nh_interface,
                            "fec": None,
                            "nh_ip": next_hop.nh_ip,
                            "rd": None,
                            "aggregate": aggregate,
                        },
                    )

        return output

    def _parse_mpls_label(self, label_str) -> Union[List[int], None]:
        """
        Parses mpls-label parameter into a list of integers
        """
        if not label_str:
            return None

        m = re.match(r"Push (?P<inner>\d+)(, Push (?P<outer>\d+))*", label_str)
        if not m:
            raise UMnetNapalmError(f"Couldn't parse label op {label_str}")

        labels = [m.group("inner")]
        if m.group("outer"):
            labels.append(m.group("outer"))

        return labels

    def _process_mpls_nh(
        self, nh_interface: str, mpls_label: str, nh_table: str
    ) -> tuple:
        """
        Parses 'nh interface', 'mpls_label', and nh_table values
        into something our standardized outpur likes.
        Returns a tuple (nh_interface:str, out_label:list, aggregate:bool )
        """
        # when the next hop is a table, not an interface we're going to
        # classify it as an aggregate label
        if not nh_interface and nh_table:
            nh_vrf = re.sub(r".inet6?.0", "", nh_table)
            return (nh_vrf, ["pop"], True)

        # aggregate labels also show up as nh_interface "lsi.X (VRF_NAME)"
        m = re.match(r"^lsi.\d+ \((\S+)\)", nh_interface)
        if m:
            return (m.group(1), [], True)

        # 'pop' label has trailing whitespace
        if mpls_label and re.match(r"^Pop", mpls_label):
            return (nh_interface, ["pop"], False)

        # 'swap' with optional 'push'
        if mpls_label:
            m = re.match(r"^Swap (?P<inner>\d+)(, Push (?P<outer>\d+))*", mpls_label)
            if m:
                out_label = [m.group("inner")]
                if m.group("outer"):
                    out_label.append(m.group("outer"))

                return (nh_interface, out_label, False)

        raise UMnetNapalmError(
            f"Cannot parse MPLS route nh: {nh_interface} label: {mpls_label}"
        )

    def get_arp_table(self, vrf="") -> ARPTableDict:
        """
        Pulls ARP table - napalm does have get_arp_table for junos but it doesn't
        support VRFs :(

        Because we have to dynamically query every VPN on the device we can't use
        junos views for this since they're based on statically defined rpc calls.
        Instead we'll have to do our own rpc calls under the hood.
        """

        # figure out the ARP aging timer - no operational command appears to show this,
        # pulling it from the config.
        if not(self._config):
            self.get_config()
        m = re.search(r"system \{.+arp {\+aging-timer (\d+);", self._config["running"])
        if m:
            self.arp_aging_timer = int(m.group(1))

        # building the rpc call as an lxml element so we can dynamically add a vrf if we want
        # if no vrf is provided all ARP entries are returned
        arp_rpc = etree.Element("get-arp-table-information")
        etree.SubElement(arp_rpc, "no-resolve")
        etree.SubElement(arp_rpc, "expiration-time")
        if vrf:
            xml_vrf = etree.SubElement(arp_rpc, "vpn")
            xml_vrf.text = vrf

        rpc_result = self.device.execute(arp_rpc)

        results = []
        for xml_entry in rpc_result.iterfind("arp-table-entry"):

            interface = xml_entry.find("interface-name").text.strip("\n")

            # MX and EX9200 include the physical interface in brackets - stripping
            # that out so only the logical interface is shown
            interface = re.sub(r" \[.+\]$","",interface)

            exp = xml_entry.find("time-to-expire")
            if exp is not None:
                age = self.arp_aging_timer - int(exp.text)
            else:
                age = 0

            results.append({
                'interface': interface,
                'ip': xml_entry.find("ip-address").text.strip("\n"),
                'mac': xml_entry.find("mac-address").text.strip("\n").upper(),
                "age": age,
                })
            
        return results




import xml.etree.ElementTree as ET
from typing import List
from datetime import timedelta
from ipaddress import ip_address, ip_network
import re

import xmltodict

from napalm_panos.panos import PANOSDriver

from umnet_napalm.models import IPInterfaceDict, RouteDict, InventoryDict
from ..base import UMnetNapalm, UMnetNapalmError
from ..utils import resolve_nh
from .utils import parse_system_state


PANOS_ROUTING_PROTOCOLS = {
    "H": "local",
    "C": "connected",
    "S": "static",
    "B": "BGP",
    "R": "RIP",
    "O": "OSPF",
    "Oi": "OSPF intra-area",
    "Oo": "OSPF inter-area",
    "O2": "OSPF External Type 2",
    "O1": "OSPF External Type 1",
}


class PANOS(UMnetNapalm, PANOSDriver):
    """
    PANOS Parser
    """

    _effective_running_config = None
    _system_state = None

    def _config_search(self, xpath: str, no_cache: bool = False) -> list:
        """
        Does an xpath search of the effective running config and returns the results.
        """
        cfg = self._get_effective_running_config(no_cache=no_cache)
        return cfg.findall(xpath)

    def _get_effective_running_config(self, no_cache: bool = False) -> ET:
        """
        Gets effective running config and saves it to the local object
        """
        if self._effective_running_config is None or no_cache:
            result = self._send_command(
                "<show><config><effective-running></effective-running></config></show>"
            )
            self._effective_running_config = ET.fromstring(result)

        return self._effective_running_config

    def _get_system_state(self, no_cache: bool = False) -> dict:
        """
        runs 'show system state' and parses the returned pseudo-json into structured output
        """
        if not self._system_state or no_cache:
            raw_result = self._send_command("<show><system><state/></system></show>")
            result = xmltodict.parse(raw_result)["response"]["result"]
            self._system_state = parse_system_state(result)

        return self._system_state

    def _get_mtu_and_comment(self, interface: ET) -> dict:
        """
        Looks up mtu and comment on a particular interface or subinterface
        object. Returns a dict with mtu and comment key/value pairs
        """
        results = {"mtu": 0, "comment": ""}

        # subinterfaces have mtu set at their same level.
        # for non-subinterfaces it is set under "layer3"
        if "." in interface.attrib["name"]:
            mtu = interface.find("./mtu")
        else:
            mtu = interface.find("./layer3/mtu")

        if mtu is not None:
            results["mtu"] = mtu.text

        comment = interface.find("./comment")
        if comment is not None:
            results["comment"] = comment.text

        return results

    def _get_interface_mtus_and_comments(self) -> dict:
        """
        Looks through the effective running config at network/interface for MTU configurations
        and descriptions, aka "comments"
        returns a dict keyed on interface names, with
        dict vaules, eg results[interface_name] = {'mtu':mtu, and 'comment': comment }
        """

        interfaces = self._config_search(".//network/interface/ethernet/entry")

        # for logical interfaces we only care about the subints
        interfaces.extend(
            self._config_search(".//network/interface/tunnel/units/entry")
        )
        interfaces.extend(
            self._config_search(".//network/interface/loopback/units/entry")
        )

        results = {}
        # looping over all the ethernet interfaces
        for interface in interfaces:
            results[interface.attrib["name"]] = self._get_mtu_and_comment(interface)

            # subinterfaces
            for subint in interface.findall("./layer3/units/entry"):
                results[subint.attrib["name"]] = self._get_mtu_and_comment(subint)

        return results

    def _send_command(self, cmd: str) -> str:
        """
        Send XML command to PANOS and get result
        """
        self.device.op(cmd=cmd)
        return self.device.xml_root()

    def get_inventory(self) -> List[InventoryDict]:
        """
        Digs through "show chassis inventory" and "show system state" to pull out linecard and optics information
        """
        # need access to a chassis nexus to implement "show chassis inventory"
        # try:
        #    result = self._send_command("<show><chassis><inventory/></chassis></show>")
        #    inventory = xmltodict.parse(result)["response"]["result"]
        # except PanXapiError:
        #    inventory = []

        # optics data is in system state
        raw_result = self._send_command(
            "<show><system><state><filter>sys.s*.p*.phy</filter></state></system></show>"
        )
        data = parse_system_state(raw_result)

        output = []
        for phy_name, phy_data in data.items():

            # skip copper and unpopulated ports
            if phy_data["media"] in ["CAT5", "SFP-Empty"]:
                continue

            # the 'phy name' is in the format 'sys.sX.pY.phy' which translates to 'EthernetX/Y'
            m = re.match(r"sys.s(\d+).y(\d+)/.phy$", phy_name)
            i_name = f"Ethernet{m.group(1)}/{m.group(2)}"

            output.append(
                {
                    "type": "optic",
                    "name": i_name,
                    "part_number": phy_data.get("vendor-name", ""),
                    "serial_number": phy_data.get("vendor-part-number", ""),
                    "parent": "",
                }
            )

        return output

    def get_ip_interfaces(self) -> List[IPInterfaceDict]:
        """
        napalm_panos doesn't support "get_routing_instances".
        While we could implement that function and combine it with "get_interfaces_ip" and
        "get_interfaces", it's simpler to just run "show interface all" to get most of what we need.

        Note that short of doing "show interface [x]" for every interface, I don't see a way
        to get MTU, so for now it's getting set to 0.
        """

        # "show interface all" gives us most of what we need here
        result = self._send_command("<show><interface>all</interface></show>")
        interfaces = xmltodict.parse(result)["response"]["result"]

        hw_interfaces_by_name = {i["name"]: i for i in interfaces["hw"]["entry"]}
        mtus_and_comments = self._get_interface_mtus_and_comments()

        l3_interfaces = []
        for l3_i in interfaces["ifnet"]["entry"]:
            if l3_i["ip"] == "N/A" and not l3_i["addr6"]:
                continue

            # vrf (aka route-domain) is stored in the 'fwd' attribute
            vrf = None
            if l3_i["fwd"] != "N/A":
                vrf = l3_i["fwd"].replace("vr:", "")

            # look up admin/oper status in 'hw' section of the output
            if "." in l3_i["name"]:
                phy_name = l3_i["name"].split(".")[0]
            else:
                phy_name = l3_i["name"]
            phy = hw_interfaces_by_name.get(phy_name, None)

            # description and mtu came from an xpath query
            mtu_and_desc = mtus_and_comments.get(
                l3_i["name"], {"mtu": 0, "comment": ""}
            )

            # all the non-IP attributes for each IP entry are the same
            l3i_entry = {
                "interface": l3_i["name"],
                "description": mtu_and_desc["comment"],
                "mtu": mtu_and_desc["mtu"] if mtu_and_desc["mtu"] else 1500,
                "admin_up": bool(phy["mode"] != "(power-down)") if phy else None,
                "oper_up": bool(phy["state"] == "up") if phy else None,
                "vrf": vrf,
            }

            # IPv4 interfaces
            if l3_i["ip"] and l3_i["ip"] != "N/A":
                ipv4_entry = l3i_entry.copy()
                ipv4_entry["ip_address"] = l3_i["ip"]
                l3_interfaces.append(ipv4_entry)

            # IPv6 interfaces are stored as a list, or a string if there's only one
            if l3_i["addr6"]:
                if isinstance(l3_i["addr6"]["member"], str):
                    l3_i["addr6"]["member"] = [l3_i["addr6"]["member"]]

                for ipv6_addr in l3_i["addr6"]["member"]:
                    ipv6_entry = l3i_entry.copy()
                    ipv6_entry["ip_address"] = ipv6_addr
                    l3_interfaces.append(ipv6_entry)

        return l3_interfaces

    def get_active_routes(self) -> List[RouteDict]:
        """
        Gets active routes from all vsyses on the PAN
        """
        result = self._send_command("<show><routing><route/></routing></show>")
        routing = xmltodict.parse(result)

        parsed_routes = []
        for route in routing["response"]["result"]["entry"]:

            # if route isn't active (no A in protocol field), skip it
            if not (route["flags"].startswith("A")):
                continue

            # look at the third character in the flags field to determine
            # protocol - don't really care about the different OSPF types
            if route["flags"][2] not in PANOS_ROUTING_PROTOCOLS:
                raise UMnetNapalmError(
                    f"{self.hostname}: Unknown panos routing protocol {route['flags']}"
                )
            protocol = PANOS_ROUTING_PROTOCOLS[route["flags"][2]]

            # next hops of all zeros indicate local routes
            if route["nexthop"] in ["0.0.0.0", "::"]:
                learned_from = "self"
            else:
                learned_from = route["nexthop"]

            # if this is a local route and we don't have a nh interface, the
            # next hop interface is also 'self' (because it's a host route or a locally-owned IP)
            if learned_from == "self" and not route["interface"]:
                route["interface"] = "self"

            parsed_routes.append(
                {
                    "vrf": route["virtual-router"],
                    "prefix": ip_network(route["destination"]),
                    "nh_interface": route["interface"],
                    "learned_from": learned_from,
                    "protocol": protocol,
                    "age": (
                        timedelta(seconds=int(route["age"])) if route["age"] else None
                    ),
                    "nh_ip": (
                        None if learned_from == "self" else ip_address(route["nexthop"])
                    ),
                    "mpls_label": None,
                    "vxlan_vni": None,
                    "vxlan_endpoint": None,
                }
            )

        # second pass to resolve next hop interface
        output = []
        for route in parsed_routes:

            if route["nh_interface"]:
                output.append(route.copy())
                continue

            nh_routes = resolve_nh(route["nh_ip"], route["vrf"], parsed_routes)
            for nh_route in nh_routes:
                resolved_route = route.copy()
                resolved_route["nh_interface"] = nh_route["nh_interface"]
                output.append(resolved_route)

        return output



import re
from typing import List, Dict
from ipaddress import ip_network
import logging

from napalm.ios import IOSDriver
from napalm.base.helpers import textfsm_extractor

from ..models import RouteDict, MPLSDict, InventoryDict
from ..base import UMnetNapalmError, UMnetNapalm
from ..utils import age_to_datetime, str_to_type, resolve_nh

IOS_PROTOCOLS = {
    "L": "local",
    "C": "connected",
    "S": "static",
    "i": "ISIS",
    "B": "BGP",
    "O": "OSPF",
}

IOS_LABEL_VALUES = {
    "": [],
    "No Label": [],
    "Pop Label": ["pop"],
}

IOS_MPLS_INTERFACES = {"Vl": "Vlan"}

logging.getLogger(__name__)


class IOS(UMnetNapalm, IOSDriver):
    """
    IOS Parser
    """

    LABEL_VALUE_MAP = {
        "": [],
        "No Label": [],
        "Pop Label": ["pop"],
    }

    INVENTORY_TO_TYPE = {
        r"Supervisor": "re",
        r"WS-": "linecard",
        r"Transceiver": "optic",
        r"Fan Module": "fan",
        r"[Pp]ower [Ss]upply": "psu",
        r"Uplink Module": "uplink-module",
        r"^Switch \d+$": "stack-member",
        r"StackPort": "optic",
        # on catalyst the optics show up as interface names
        # sometimes the names are abbreviated and sometimes they
        # are not (sigh)
        r"^(Twe|Te|Gi|Two).+\d$": "optic",
        # don't care about these inventory items
        r"(Clock FRU|Daughterboard|Feature Card|Forwarding Card)": None,
        # this is the chassis id of the master of the stack,
        # we
        r"Stack": None,
    }

    def _parse_protocol_abbr(self, abbr) -> str:
        if abbr in IOS_PROTOCOLS:
            return IOS_PROTOCOLS[abbr]

        raise UMnetNapalmError(f"Unknown IOS protocol abbr {abbr}")

    def _parse_label_value(self, label) -> list:
        """
        Parses mpls label value into normalized data
        """
        if label in IOS_LABEL_VALUES:
            return IOS_LABEL_VALUES[label]

        return [label]

    def _get_route_labels(self) -> Dict[tuple, str]:
        """
        Runs "show bgp vpnv4 unicast labels" and parses the result.

        The output is a dictionary with (vrf, prefix) as key
        and the outbound label as a value, eg
        output[ ("vrf_VOIP_NGFW", "0.0.0.0/0") ] = "12345"
        """

        raw_labels = self._send_command("show bgp vpnv4 unicast all labels")
        labels = textfsm_extractor(self, "sh_bgp_vpnv4_unicast_all_labels", raw_labels)

        # default route shows up as '0.0.0.0' so we have to munge that
        output = {}
        for l in labels:
            prefix = "0.0.0.0/0" if l["prefix"] == "0.0.0.0" else l["prefix"]
            output[(l["vrf"], ip_network(prefix))] = l["out_label"]

        return output

    def get_active_routes(self) -> List[RouteDict]:
        """
        Parses "show ip route vrf *" for IOS. Will also run
        "show bgp vpnv4 unicast labels" to get label bindings
        """

        parsed_routes = []
        output = []

        raw_routes = self._send_command("show ip route vrf *")
        routes = textfsm_extractor(self, "sh_ip_route_vrf_all", raw_routes)

        for route in routes:
            protocol = self._parse_protocol_abbr(route["proto_1"])

            # "learned from" is one of the keys in our route table that determines
            # uniqueness, as such we need to make sure it's set. Usually it's
            # the IP of the advertising router, but for local/direct/static it should
            # get set to 'self'
            if route["nh_ip"]:
                learned_from = route["nh_ip"]
            elif protocol in ["local", "connected", "static"]:
                learned_from = "self"
            else:
                raise UMnetNapalmError(f"Could not determine learned from for {route}")

            parsed_routes.append(
                {
                    "vrf": route["vrf"] if route["vrf"] else "default",
                    "prefix": str_to_type(route["prefix"]),
                    "nh_interface": route["nh_interface"],
                    "learned_from": learned_from,
                    "protocol": self._parse_protocol_abbr(route["proto_1"]),
                    "age": age_to_datetime(route["age"]),
                    "nh_ip": str_to_type(route["nh_ip"]),
                    "mpls_label": None,
                    "vxlan_vni": None,
                    "vxlan_endpoint": None,
                }
            )

        # Second pass to resolve recursive routes
        labels = self._get_route_labels()
        for route in parsed_routes:
            # routes that were already resolved can be skipped
            if route["nh_interface"]:
                output.append(route)
                continue

            # if we found an integer label mapping for this route we will save the label.
            # note that the way our old core is architected, there is only ever one label
            # in the stack
            if re.match(r"^\d+$", labels.get((route["vrf"], route["prefix"]), "")):
                route["mpls_label"] = [int(labels[(route["vrf"], route["prefix"])])]
                # also making an assumption that we need to resolve the next hop in
                # our default table
                nh_table = "default"

            # if we didn't find a label, assume this is vrf lite and the next hop
            # is in our local vrf
            else:
                nh_table = route["vrf"]

            # attempt to resolve next hop
            nh_lpms = resolve_nh(route["nh_ip"], nh_table, parsed_routes)
            if not nh_lpms:
                logging.error(f"Could not resolve next hop for {route}")
                continue

            # for every lpm found, create an entry for this recursive route
            for lpm in nh_lpms:
                resolved = route.copy()
                resolved["nh_ip"] = lpm["nh_ip"]
                resolved["nh_interface"] = lpm["nh_interface"]
                output.append(resolved)

        return output

    def get_inventory(self) -> list[InventoryDict]:
        """
        Parses "show inventory" for IOS
        """
        raw_inventory = self._send_command("show inventory")
        inventory = textfsm_extractor(self, "sh_inventory", raw_inventory)

        output = []
        for entry in inventory:
            inventory_type = self._get_inventory_type(entry["name"])

            if not inventory_type:
                continue

            output.append(
                {
                    "type": inventory_type,
                    "name": entry["name"],
                    "part_number": entry["pid"],
                    "serial_number": entry["sn"],
                    "parent": None,
                }
            )

        return output

    def get_mpls_switching(self) -> List[MPLSDict]:
        """
        Parses "show mpls forwarding table" for IOS
        """
        raw_labels = self._send_command("show mpls forwarding-table")
        labels = textfsm_extractor(self, "sh_mpls_forwarding_table", raw_labels)

        output = []
        for entry in labels:
            # extract RD from 'FEC'
            m = re.match(r"([\d\.]+:\d+):(\d+.\d+.\d+.\d+\/\d+)", entry["fec"])
            if m:
                rd = m.group(1)
                fec = m.group(2)
            else:
                rd = None
                fec = entry["fec"]

            aggregate = bool(entry["vrf"])
            nh_interface = entry["vrf"] if entry["vrf"] else entry["nh_interface"]

            output.append(
                {
                    "in_label": str_to_type(entry["in_label"]),
                    "fec": str_to_type(fec),
                    "out_label": self._parse_label_value(entry["out_label"]),
                    "nh_ip": str_to_type(entry["nh_ip"]),
                    "nh_interface": str_to_type(nh_interface),
                    "rd": rd,
                    "aggregate": aggregate,
                }
            )

        return output



from .ios import IOS
from .nxos import NXOS
from .junos import Junos, junos_views
from .abstract_base import AbstractUMnetNapalm


class FixtureNotFoundError(FileNotFoundError):
    pass


class MockParser(AbstractUMnetNapalm):
    """
    Overloads the _send_command methods
    in the different parsers so that the data comes from
    a file instead of a live device
    """

    fixture_vendor = "mock"

    def __init__(self, fixture_path: str, hostname: str):
        self.fixture_path = fixture_path
        self.hostname = hostname

    def _send_command(self, command: str):
        """
        Retrieves command response from a file instead of a live device
        File must be saved at the path provided to the mock constructor, named
        based on the command with underscores instead of spaces.

        Note that for Junos the "command" is actually the junos view
        """
        fixture = command.replace(" ", "_")
        fixture_file = (
            f"{self.fixture_path}/{self.fixture_vendor}/{self.hostname}/{fixture}.txt"
        )
        print(fixture_file)
        try:
            with open(fixture_file, encoding="utf-8") as fh:
                result = fh.read()
        except FileNotFoundError as e:
            raise FixtureNotFoundError from e

        return result


class MockIOS(MockParser, IOS):
    """
    Mock class for testing the IOS parser.
    """

    fixture_vendor = "ios"


class MockNXOS(MockParser, NXOS):
    """
    Mock class for testing the NXOS parser.
    """

    fixture_vendor = "nxos"


class MockJunos(MockParser, Junos):
    """
    Mock class for testing the Junos parser.
    """

    fixture_vendor = "junos"

    def _send_command(self, command: str):
        """
        You can pass a "path" attribute to the pyez View constructor
        instead of a device and it will read the data from a file - woot.
        """

        view = getattr(junos_views, command)
        fixture_file = f"{self.fixture_path}/{self.fixture_vendor}/{self.hostname}/{view.GET_RPC}.xml"
        result = view(path=fixture_file)  # pylint: disable=no-member
        return result.get()



from ipaddress import ip_network, ip_interface, ip_address
from typing import Optional

from datetime import timedelta

from typing_extensions import TypedDict


# using TypedDict to model our standardized output because that's
# what nampalm does and I like it - ref
# https://github.com/napalm-automation/napalm/blob/develop/napalm/base/models.py
VNIDict = TypedDict(
    "VNIDict",
    {
        "vni": int,
        "mcast_group": Optional[ip_address],
        "vrf": Optional[str],
        "vlan_id": Optional[int],
    },
)

RouteDict = TypedDict(
    "RouteDict",
    {
        "vrf": str,
        "prefix": ip_network,
        "nh_interface": str,
        "learned_from": str,
        "protocol": str,
        "age": timedelta,
        "nh_ip": Optional[ip_address],
        "mpls_label": Optional[list[str]],
        "vxlan_vni": Optional[int],
        "vxlan_endpoint": Optional[ip_address],
    },
)

MPLSDict = TypedDict(
    "MPLSDict",
    {
        "in_label": str,
        "out_label": list,
        "nh_interface": Optional[str],
        "fec": Optional[ip_network],
        "nh_ip": Optional[ip_address],
        "rd": Optional[str],
        "aggregate": bool,
    },
)

IPInterfaceDict = TypedDict(
    "IPInterfaceDict",
    {
        "ip_address": ip_interface,
        "interface": str,
        "description": str,
        "mtu": int,
        "admin_up": bool,
        "oper_up": bool,
        "vrf": str,
    },
)

VersionDict = TypedDict(
    "VersionDict",
    {
        "ip": ip_address,
        "version": str,
        "vendor": str,
        "model": str,
        "serial": str,
    },
)

NeighborsDict = TypedDict(
    "NeighborsDict",
    {
        "port": str,
        "remote_device": str,
        "remote_port": str,
    },
)

InventoryDict = TypedDict(
    "InventoryDict",
    {
        "type": str,
        "name": str,
        "part_number": str,
        "serial_number": str,
        "parent": str,
    },
)



from datetime import timedelta
import re
from ipaddress import ip_network, ip_interface, ip_address
from ipaddress import (
    IPv4Address,
    IPv6Address,
    IPv4Interface,
    IPv6Interface,
    IPv4Network,
    IPv6Network,
)
import json
from typing import List
from .models import RouteDict


# eventually want to pull this from a more central place
# as we've defined it in so many places already
INTERFACE_ABBRS = {
    # ios
    "Fastethernet": "Fa",
    "GigabitEthernet": "Gi",
    "TwoGigabitEthernet": "Tw",
    "TenGigabitEthernet": "Te",
    "TwentyFiveGigE": "Twe",
    "Port-channel": "Po",
    "Loopback": "Lo",
    # nxos
    "Ethernet": "Eth",
    "port-channel": "Po",
    "loopback": "Lo",
}


class UMnetNapalmJsonEncoder(json.JSONEncoder):
    """
    custom json encoder that handles types
    in our results that aren't encodable by default.
    Reference the "Extending JSONEncoder" section of: https://docs.python.org/3/library/json.html

    Currently ipaddress and timedelta types are the only non-encodable
    types among the results
    """

    def default(self, obj):
        if isinstance(
            obj,
            (
                IPv4Address,
                IPv6Address,
                IPv4Interface,
                IPv6Interface,
                IPv4Network,
                IPv6Network,
                timedelta,
            ),
        ):
            return str(obj)

        return super().default(obj)


def abbr_interface(interface: str) -> str:
    """
    Converts long version of interface name to short one
    if applicable
    """
    for long, short in INTERFACE_ABBRS.items():
        if interface.startswith(long):
            return interface.replace(long, short)

    return interface


def age_to_datetime(age: str) -> timedelta:
    """
    Across platforms age strings can be:
    10y5w, 5w4d, 05d04h, 01:10:12, 3w4d 01:02:03
    """
    days = 0
    hours = 0
    minutes = 0
    seconds = 0
    m = re.search(
        r"^((?P<years>\d+)y)*((?P<weeks>\d+)w)*((?P<days>\d+)d)*((?P<hours>\d+)h)*",
        age,
    )
    if m:
        if m.group("years"):
            days += int(m.group("years")) * 365
        if m.group("weeks"):
            days += int(m.group("weeks")) * 7
        if m.group("days"):
            days += int(m.group("days"))
        if m.group("hours"):
            hours += int(m.group("hours"))

    m = re.search(r"(?P<hours>\d+):(?P<minutes>\d+):(?P<seconds>\d+)$", age)
    if m:
        if m.group("hours"):
            hours += int(m.group("hours"))
        if m.group("minutes"):
            minutes += int(m.group("minutes"))
        if m.group("seconds"):
            seconds += int(m.group("seconds"))

    return timedelta(
        days=days,
        hours=hours,
        minutes=minutes,
        seconds=seconds,
    )


def resolve_nh(nh_ip: str, nh_table: str, routes: List[RouteDict]) -> List[RouteDict]:
    """
    Given a next hop IP and next hop table, search a list of
    routes for that next hop and return LPMs that have a next hop
    interface
    """
    nh_routes = [
        r
        for r in routes
        if r["vrf"] == nh_table
        and r["prefix"].version == ip_network(nh_ip).version
        and r["prefix"].supernet_of(ip_network(nh_ip))
        and r["nh_interface"]
    ]

    if not nh_routes:
        return []

    nh_routes.sort(key=lambda x: x["prefix"].prefixlen, reverse=True)

    # the first entry is a longest prefix match because of our sort,
    # but with ECMP there could be more than one!
    # we'll peel routes off the front with matching lengths to find
    # the rest.
    lpms = [nh_routes[0]]
    if len(nh_routes) > 1:
        for nh_route in nh_routes[1::]:
            if nh_route["prefix"].prefixlen == nh_routes[0]["prefix"].prefixlen:
                lpms.append(nh_route)

    return lpms


def str_to_type(string: str) -> object:
    """
    Converts string output to obvious types.
    empty quote is converted to "None".
    ip_address, ip_interface, and ip_network are converted to their appropriate objecst
    integers are converted to int
    """
    if string == "":
        return None
    if re.match(r"^\d+$", string):
        return int(string)

    for ip_type in [ip_network, ip_interface, ip_address]:
        try:
            return ip_type(string)
        except ValueError:
            pass

    return string



from typing import List
from ipaddress import ip_interface, ip_network, ip_address
from copy import deepcopy
import re

from napalm.nxos_ssh import NXOSSSHDriver
from napalm.base.helpers import textfsm_extractor
from napalm.base.models import FactsDict

from ..base import UMnetNapalmError, UMnetNapalm
from ..models import RouteDict, MPLSDict, IPInterfaceDict, VNIDict, InventoryDict
from ..utils import age_to_datetime, str_to_type, resolve_nh, abbr_interface


class NXOS(UMnetNapalm, NXOSSSHDriver):
    """
    NXOS Parser
    """

    LABEL_VALUE_MAP = {
        "No": [],
        "Pop": ["pop"],
    }

    # for nexus we're going to map the 'description' provided by
    # show inventory to the type
    INVENTORY_TO_TYPE = {
        # note we're using the fact that this dict gets evaluated
        # sequentially
        r"Fabric Module$": "fabric-module",
        r"Supervisor Module$": "re",
        r"Fan Module$": "fan",
        r"Module$": "linecard",
        r"Power Supply$": "psu",
        r"^Transceiver": "optic",
        # don't care about chassis or system controller types
        r"Chassis": None,
        r"System Controller": None,
    }

    def _get_nxos_inventory_type(self, name: str, desc: str) -> str:
        """
        Figures out the inventory type based on the 'name' and 'desc'
        fields from 'show inventory all'
        """

        # the description field is all over the place for 3rd party
        # optics, so we want to rely on the 'name' field, which
        # is always the interface name
        if name.startswith("Ethernet"):
            return "optic"

        # otherwise we want to figure this out based on description
        return self._get_inventory_type(desc)

    def get_active_routes(self) -> List[RouteDict]:
        """
        Parses 'sh ip route detail vrf all'
        """

        parsed_routes = []
        raw_routes = self._send_command("show ip route detail vrf all")
        routes = textfsm_extractor(self, "sh_ip_route_detail_vrf_all", raw_routes)

        for route in routes:
            # skip 'broadcast' and 'local' routes, we don't really care about these
            if route["protocol"] in ["broadcast", "local"]:
                continue

            # "learned from" is one of the keys in our route table that determines
            # uniqueness, as such we need to make sure it's set. Usually it's
            # the IP of the advertising router, but for local/direct/static it should
            # get set to 'self'
            if route["nh_ip"]:
                learned_from = route["nh_ip"]
            elif route["protocol"] in ["direct", "local", "vrrp", "static"]:
                learned_from = "self"
            else:
                raise UMnetNapalmError(f"Could not determine learned from for {route}")

            parsed_routes.append(
                {
                    "vrf": route["vrf"],
                    "prefix": ip_network(route["prefix"]),
                    "nh_interface": route["nh_interface"],
                    "learned_from": learned_from,
                    "protocol": route["protocol"],
                    "age": age_to_datetime(route["age"]),
                    "nh_ip": str_to_type(route["nh_ip"]),
                    "mpls_label": [route["label"]] if route["label"] else [],
                    "vxlan_vni": str_to_type(route["vni"]),
                    "vxlan_endpoint": str_to_type(route["nh_ip"]),
                }
            )

        # second pass to deal with recursive routes
        output = []
        for route in parsed_routes:
            if route["nh_interface"]:
                output.append(route)
                continue

            # attempt to resolve next hop
            nh_table = (
                "default" if route["mpls_label"] or route["vxlan_vni"] else route["vrf"]
            )
            nh_lpms = resolve_nh(route["nh_ip"], nh_table, parsed_routes)
            if not nh_lpms:
                self.log.error(f"Could not resolve next hop for {route}")
                continue
            # for every lpm found create an entry in our table for this recursive
            # route
            for lpm in nh_lpms:
                resolved = deepcopy(route)
                resolved["nh_ip"] = lpm["nh_ip"]
                resolved["nh_interface"] = lpm["nh_interface"]
                resolved["mpls_label"].extend(lpm["mpls_label"])
                output.append(resolved)

        return output

    def get_facts(self) -> FactsDict:
        """
        Cleans up model number on napalm get_facts
        """

        results = super().get_facts()

        model = results["model"]
        m = re.match(r"Nexus(3|9)\d+ (\S+) (\(\d Slot\) )*Chassis$", model)

        # some models have the "N9K" or "N3K already in them, some don't.
        if m and re.match(r"N\dK", m.group(2)):
            results["model"] = m.group(2)
        elif m:
            results["model"] = f"N{m.group(1)}K-{m.group(2)}"

        return results

    def get_mpls_switching(self) -> List[MPLSDict]:
        """
        parses show mpls into a dict that outputs
        aggregate labels and
        """
        output = []

        raw_entries = self._send_command("show mpls switching detail")
        entries = textfsm_extractor(self, "sh_mpls_switching_detail", raw_entries)

        for entry in entries:
            # for aggregate labels the next hop is the VRF
            nh_interface = entry["vrf"] if entry["vrf"] else entry["nh_interface"]
            output.append(
                {
                    "in_label": str_to_type(entry["in_label"]),
                    "out_label": self._parse_label_value(entry["out_label"]),
                    "fec": str_to_type(entry["fec"]),
                    "nh_ip": str_to_type(entry["nh_ip"]),
                    "nh_interface": str_to_type(nh_interface),
                    "rd": str_to_type(entry["rd"]),
                    "aggregate": bool(entry["vrf"]),
                }
            )

        return output

    def get_inventory(self) -> List[InventoryDict]:
        """
        Parses "show inventory"
        """
        raw_inventory = self._send_command("show inventory all")
        inventory = textfsm_extractor(self, "sh_inventory_all", raw_inventory)

        output = []
        for entry in inventory:

            # removing quotes from name and description fields,
            # which are enquoted for certain types
            for key in ["name", "desc"]:
                entry[key] = entry[key].replace('"', "")

            inventory_type = self._get_nxos_inventory_type(entry["name"], entry["desc"])
            if not inventory_type:
                continue

            # if this is a linecard in slot 1 and the model number
            # doesn't look like a linecard, then this is a fixed config device
            # and the "lincard" is really the chassis and we want to ingore it
            if (
                inventory_type == "linecard"
                and entry["name"] == "Slot 1"
                and re.match(r"N[39]K-C", entry["pid"])
            ):
                continue

            output.append(
                {
                    "type": inventory_type,
                    "name": entry["name"],
                    "part_number": entry["pid"],
                    "serial_number": entry["sn"],
                    "parent": None,
                }
            )

        return output

    def get_ip_interfaces(self) -> List[IPInterfaceDict]:
        output = []

        # NOTE: because "get_interfaces" currently has a bug for loopback and mgmt IPs on Nexus,
        # we're going to do our own parsing. WE can actually get everything we need
        # with a combination of "show ip int vrf all" and the napalm get_interfaces
        # function
        raw_interfaces = self._send_command("show ip interface vrf all")
        ip_interfaces = textfsm_extractor(
            self, "sh_ip_interface_vrf_all", raw_interfaces
        )
        phy_interfaces = super().get_interfaces()

        for i in ip_interfaces:
            # some l3 interfaces don't have IP addresses (like EVPN L3VPN vrf SVIs)
            # we want to skip those.
            if not i["ip_address"]:
                continue
            phy_i = phy_interfaces.get(i["interface"], {})
            output.append(
                {
                    "ip_address": ip_interface(f'{i["ip_address"]}/{i["prefixlen"]}'),
                    "interface": abbr_interface(i["interface"]),
                    "description": phy_i.get("description", ""),
                    "mtu": int(i["mtu"]),
                    "admin_up": (i["admin_state"] == "admin-up"),
                    "oper_up": (i["protocol_state"] == "protocol-up"),
                    "vrf": i["vrf"],
                }
            )

        return output

    def get_vni_information(self) -> List[VNIDict]:
        """
        Runs "show nve vni" to get vni info
        """
        output = []
        raw_vnis = self._send_command("show nve vni")
        vnis = textfsm_extractor(self, "sh_nve_vni", raw_vnis)

        for vni in vnis:

            output.append(
                {
                    "vni": str_to_type(vni["vni"]),
                    "mcast_group": (
                        None
                        if vni["mcast_grp"] == "n/a"
                        else ip_address(vni["mcast_grp"])
                    ),
                    "vrf": vni["bd_vrf"] if vni["type"] == "L3" else None,
                    "vlan_id": (
                        vni["bd_vrf"]
                        if vni["type"] == "L2" and vni["bd_vrf"] != "UC"
                        else None
                    ),
                },
            )

        return output



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['umnet_napalm',
 'umnet_napalm.ios',
 'umnet_napalm.junos',
 'umnet_napalm.nxos',
 'umnet_napalm.panos']

package_data = \
{'': ['*'],
 'umnet_napalm.ios': ['utils/textfsm_templates/*'],
 'umnet_napalm.nxos': ['utils/textfsm_templates/*']}

install_requires = \
['napalm-panos>=0.6.2,<0.7.0', 'napalm>=4.0.0,<5.0.0']

setup_kwargs = {
    'name': 'umnet-napalm',
    'version': '0.1.1',
    'description': 'A custom version of NAPALM for UMnet',
    'long_description': '# umnet-napalm\nThis is a project that augments the [NAPALM](https://napalm.readthedocs.io/en/latest/) library in ways that are relevant to our interests.\nMore specifically, new [getter functions](https://napalm.readthedocs.io/en/latest/support/index.html#getters-support-matrix) have been implemented to pull\ndata from routers and parse it into a vender agnostic format.\n\n\nCurrently, the following NAPALM classes have been augmented:\n* IOS\n* NXOS-SSH\n* Junos\n* PANOS (from [napalm-panos](https://github.com/napalm-automation-community/napalm-panos))\n\nThe following "getter" methods have been added:\n* `get_ip_interfaces`\n* `get_active_routes`\n* `get_mpls_switching`\n* `get_vni_information`\n',
    'author': 'Amy Liebowitz',
    'author_email': 'amylieb@umich.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.12',
}


setup(**setup_kwargs)



from typing import List
import re
from ipaddress import ip_interface
import logging
from napalm.base import NetworkDriver

from .models import IPInterfaceDict, RouteDict, MPLSDict, VNIDict
from .utils import abbr_interface

# from .abstract_base import AbstractUMnetNapalm


class UMnetNapalmError(Exception):
    """
    Generic error class
    """


class UMnetNapalm(NetworkDriver):
    """
    Base class for um-specific (but also non-vendor specific)
    implementations.
    """

    log = logging.getLogger("nornir")

    # populate in child classes
    IGNORE_INTERFACES = []
    LABEL_VALUE_MAP = {}
    INVENTORY_TO_TYPE = {}

    def _get_inventory_type(self, name: str) -> str:
        """
        Maps the name of the inventory item to its type
        """
        for pattern, inv_type in self.INVENTORY_TO_TYPE.items():
            if re.search(pattern, name):
                return inv_type

        raise UMnetNapalmError(f"{self.hostname}: Unknown inventory item {name}")

    def get_ip_interfaces(self) -> List[IPInterfaceDict]:
        """
        Uses a series napalm getters to retrieve layer 3 ip address information.

        getter models reference:
        https://github.com/napalm-automation/napalm/blob/develop/napalm/base/base.py
        """
        phy_interfaces = self.get_interfaces()
        ip_interfaces = self.get_interfaces_ip()
        vrfs = self.get_network_instances()

        # rekeying the vrf data by interface to make our life easier
        vrf_interfaces = {}
        for vrf, vrf_data in vrfs.items():
            for i in vrf_data["interfaces"]["interface"]:
                vrf_interfaces[i] = vrf

        output = []
        for i, data in ip_interfaces.items():
            if self._ignore_interface(i):
                continue

            if i not in phy_interfaces or i not in vrf_interfaces:
                continue

            phy_i = phy_interfaces[i]
            vrf_i = vrf_interfaces[i]

            all_ips = data.get("ipv4", {})
            all_ips.update(data.get("ipv6", {}))

            for a, a_data in all_ips.items():
                ip = ip_interface(f'{a}/{a_data["prefix_length"]}')

                # skip link-local addresses
                if ip.is_link_local:
                    continue

                output.append(
                    {
                        "ip_address": ip,
                        "interface": abbr_interface(i),
                        "description": phy_i["description"],
                        "mtu": phy_i["mtu"],
                        "admin_up": phy_i["is_enabled"],
                        "oper_up": phy_i["is_up"],
                        "vrf": vrf_i,
                    }
                )
        return output

    def get_active_routes(self) -> List[RouteDict]:
        """get active routes"""
        return []

    def get_mpls_switching(self) -> List[MPLSDict]:
        """get mpls switching (the mpls forwarding table)"""
        return []

    def get_vni_information(self) -> List[VNIDict]:
        """get vni to vlan and VRF mapping"""
        return []

    def _ignore_interface(self, interface: str) -> bool:
        for ignore in self.IGNORE_INTERFACES:
            if re.search(ignore, interface):
                return True
        return False

    def _parse_label_value(self, label) -> list:
        """
        Parses mpls label value into normalized data
        """
        if label in self.LABEL_VALUE_MAP:
            return self.LABEL_VALUE_MAP[label]

        return [label]
